r"""
==========================================================
Pretty-printer for quantum objects (:mod:`rqutils.qprint`)
==========================================================

.. currentmodule:: rqutils.qprint

QPrint API
==========

.. autosummary::
   :toctree: ../generated
   
   qprint
   pauliprint
"""

from typing import Tuple, List, Sequence, Union, Optional, Any
from numbers import Number
import builtins
from dataclasses import dataclass
from enum import Enum
import numpy as np

try:
    import scipy
except ImportError:
    has_scipy = False
else:
    has_scipy = True

try:
    import matplotlib as mpl
    import matplotlib.pyplot as plt
except ImportError:
    has_mpl = False
else:
    has_mpl = True

    MATPLOTLIB_INLINE_BACKENDS = {
        "module://ipykernel.pylab.backend_inline",
        "module://matplotlib_inline.backend_inline",
        "nbAgg",
    }
    
try:
    from qutip import Qobj
except ImportError:
    has_qutip = False
else:
    has_qutip = True
    
from . import paulis
from ._types import array_like

class LaTeXRepr:
    def __init__(self, obj):
        self.obj = obj
        
    def _repr_latex_(self):
        return self.obj.latex()

PrintReturnType = Union[str, LaTeXRepr]
if has_mpl:
    PrintReturnType = Union[PrintReturnType, mpl.figure.Figure]


class QPrintBase:
    """Helper class to compose an expression of a given quantum object.
    
    This is a base class for QPrint which performs numerical processing of the components in the input
    quantum object. Basis labeling is handled by the concrete subclasses.
    
    Args:
        qobj: Input quantum object.
        amp_norm: Specification of the normalization of amplitudes by (numeric devisor, unit in LaTeX).
        phase_norm: Specification of the normalization of phases by (numeric devisor, unit in LaTeX).
        global_phase: Specification of the phase to factor out. Give a numeric offset or 'mean'.
        terms_per_row: Number of terms to show per row.
        amp_format: Format for the numerical value of the amplitude absolute values.
        phase_format: Format for the numerical value of the phases.
        epsilon: Numerical cutoff for ignoring amplitudes (relative to max) and phase (absolute).
        lhs_label: If not None, prepend 'label = ' to the printout.
    """
    
    @dataclass
    class Term:
        index: tuple
        sign: int
        amp: str
        phase: str
        
    def __init__(
        self,
        qobj: Any,
        amp_norm: Optional[Union[Number, Tuple[Number, str]]] = None,
        phase_norm: Optional[Tuple[Number, str]] = (np.pi, 'π'),
        global_phase: Optional[Union[Number, str]] = None,
        terms_per_row: int = 0,
        amp_format: str = '.3f',
        phase_format: str = '.2f',
        epsilon: float = 1.e-6,
        lhs_label: Optional[str] = None
    ):
        self.qobj = qobj
        self.amp_norm = amp_norm
        self.phase_norm = phase_norm
        self.global_phase = global_phase
        self.terms_per_row = terms_per_row
        self.amp_format = amp_format
        self.phase_format = phase_format
        self.epsilon = epsilon
        self.lhs_label = lhs_label
        
    def _process(self, obj=None) -> Tuple[int, str, str, List[List[Term]]]:
        """Compose a list of QPrintTerms."""
        if obj is None:
            obj = self.qobj
            
        qobj, data = self._qobj_data(obj)

        # Amplitude format template
        amp_template = f'{{:{self.amp_format}}}'

        # Phase format template
        phase_template = f'{{:{self.phase_format}}}'

        ## Preprocess the qobj

        # Absolute value and phase of the amplitudes
        absamp = np.abs(data)
        phase = np.angle(data)

        # Normalize the abs amplitudes and identify integral values
        if self.amp_norm is not None:
            if isinstance(self.amp_norm, tuple):
                absamp /= self.amp_norm[0]
                global_amp = self.amp_norm[1]
            else:
                absamp /= self.amp_norm
                if np.isclose(np.round(self.amp_norm), self.amp_norm, rtol=self.epsilon):
                    global_amp = f'{np.round(self.amp_norm)}'
                else:
                    global_amp = amp_template.format(self.amp_norm)
                
        else:
            global_amp = ''
            
        rounded_amp = np.round(absamp).astype(int)
        amp_is_int = np.isclose(rounded_amp, absamp, rtol=self.epsilon)
        rounded_amp = np.where(amp_is_int, rounded_amp, -1)
        
        # Shift and normalize the phases and identify integral values
        phase_offset = 0.
        if self.global_phase is not None:
            if self.global_phase == 'mean':
                phase_offset = np.mean(phase)
            else:
                phase_offset = self.global_phase
                
            phase -= phase_offset

        twopi = 2. * np.pi
        
        while np.any((phase < 0.) | (phase >= twopi)):
            phase = np.where(phase >= 0., phase, phase + twopi)
            phase = np.where(phase < twopi, phase, phase - twopi)

        def normalize_phase(phase):
            reduced_phase = phase / (np.pi / 2.)
            axis_proj = np.round(reduced_phase).astype(int)
            on_axis = np.isclose(axis_proj, reduced_phase, rtol=0., atol=self.epsilon)
            axis_proj = np.where(on_axis, axis_proj, -1)

            if self.phase_norm is not None:
                phase /= self.phase_norm[0]

            rounded_phase = np.round(phase).astype(int)
            phase_is_int = np.isclose(rounded_phase, phase, rtol=0., atol=self.epsilon)
            rounded_phase = np.where(phase_is_int, rounded_phase, -1)
            
            return phase, axis_proj, rounded_phase
        
        def sign_and_phase(phase, axis_proj, rounded_phase):
            if axis_proj == -1:
                # Not on Re or Im axis
                if rounded_phase == -1:
                    expr = phase_template.format(phase)
                else:
                    expr = f'{rounded_phase}'
                    
                sign = 1

            else:
                if axis_proj % 2 == 1:
                    expr = '/'
                else:
                    expr = '0'

                if axis_proj >= 2:
                    sign = -1
                else:
                    sign = 1
                    
            return sign, expr
        
        norm_offset, offset_proj, rounded_offset = normalize_phase(phase_offset)
        global_sign, global_phase = sign_and_phase(norm_offset, offset_proj, rounded_offset)
        
        norm_phase, axis_proj, rounded_phase = normalize_phase(phase)

        ## Compose the terms
        
        # Show only terms with absamp < absamp * epsilon
        amp_atol = np.amax(absamp) * self.epsilon
        amp_is_zero = np.isclose(np.zeros_like(absamp), absamp, atol=amp_atol)
        term_indices = list(zip(*np.logical_not(amp_is_zero).nonzero())) # convert into list of tuples

        # List of terms
        terms = []

        for idx in term_indices:
            sign, phase_expr = sign_and_phase(norm_phase[idx], axis_proj[idx], rounded_phase[idx])
                
            if rounded_amp[idx] == -1:
                amp_expr = amp_template.format(absamp[idx])
            else:
                amp_expr = f'{rounded_amp[idx]}'
                
            terms.append(QPrintBase.Term(index=idx, sign=sign, amp=amp_expr, phase=phase_expr))

        return global_sign, global_amp, global_phase, terms
    
    def _qobj_data(self, obj=None):
        if obj is None:
            obj = self.qobj

        if has_qutip and isinstance(obj, Qobj):
            qobj = obj.data
            data = qobj.data
        elif has_scipy and isinstance(obj, scipy.sparse.csr_matrix):
            qobj = obj
            data = qobj.data
        elif isinstance(obj, np.ndarray):
            qobj = obj
            data = qobj
        else:
            raise NotImplementedError(f'qprint not implemented for {type(obj)}')
            
        return qobj, data
    
    def __repr__(self):
        expr = self._format_lhs('text')
        
        if expr:
            expr += ' = '

        pre_expr, lines = self._make_lines('text')
        
        if pre_expr:
            expr += f'{pre_expr} ('

        expr += '\n'.join(lines)
        
        if pre_expr:
            expr += ')'
            
        return expr
    
    def latex(self) -> str:
        pre_expr, lines = self._make_lines('latex')

        if pre_expr:
            lines[0] = fr' \left( {lines[0]}'
            lines[-1] += r' \right)'

            if len(lines) > 1:
                lines[0] += r' \right.'
                lines[-1] = r'\left. ' + lines[-1]
                
        if len(lines) > 1:
            lines = list(f'& {line}' for line in lines)
            
        if pre_expr:
            lines[0] = f'{pre_expr} {lines[0]}'
            
        lhs = self._format_lhs('latex')
                
        if lhs:
            lines[0] = f'{lhs} = {lines[0]}'
                
        if len(lines) == 1:
            return fr'$\displaystyle {lines[0]}$'
        else:
            return r'\begin{split} ' + r' \\ '.join(lines) + r' \end{split}'
        
    def mpl(self):
        if not has_mpl:
            raise RuntimeError('Matplotlib is not available')
        
        pre_expr, lines = self._make_lines('latex')
        
        if pre_expr:
            lines[0] = f'{pre_expr} ({lines[0]}'
            lines[-1] += ')'
            
        lhs = self._format_lhs('latex')
            
        if lhs:
            lines[0] = f'{lhs} = {lines[0]}'
            
        fig, ax = plt.subplots(1, figsize=[10., 0.5 * len(lines)])
        ax.axis('off')
    
        num_rows = len(lines)
        for irow, line in enumerate(lines):
            ax.text(0.5, 1. / num_rows * (num_rows - irow - 1), f'${line}$', fontsize='x-large', ha='right')
            
        if mpl.get_backend() in MATPLOTLIB_INLINE_BACKENDS:
            plt.close(fig)
                
        return fig
    
    def _make_lines(self, mode) -> list:
        global_sign, global_amp, global_phase, terms = self._compose()
        
        pre_expr = ''
        
        if global_sign == -1:
            pre_expr += '-'
        
        pre_expr += global_amp
        pre_expr += self._format_phase(global_phase, mode)

        lines = []
        line_expr = ''
        num_terms = 0
        
        for term in terms:
            if lines or line_expr:
                if term.sign == -1:
                    line_expr += ' - '
                else:
                    line_expr += ' + '
            elif term.sign == -1:
                line_expr += '-'
                
            if term.amp != '1':
                line_expr += term.amp
                
            line_expr += self._format_phase(term.phase, mode)
            line_expr += self._format_label(term, mode)
                
            num_terms += 1
            if num_terms == self.terms_per_row:
                lines.append(line_expr)
                line_expr = ''
                num_terms = 0
                
        if num_terms != 0:
            lines.append(line_expr)
            
        if not lines:
            lines = ['0']

        return pre_expr, lines
        
    def _format_phase(self, phase_expr, mode):
        if phase_expr == '0':
            return ''
        elif phase_expr == '/':
            return 'i'
        
        if mode == 'text':
            expr = '['

            if self.phase_norm is not None and self.phase_norm[1]:
                if phase_expr == '1':
                    expr += self.phase_norm[1]
                elif self.phase_norm[1][0].isnumeric():
                    expr += f'{phase_expr}({self.phase_norm[1]})'
                else:
                    expr += f'{phase_expr}{self.phase_norm[1]}'
            else:
                expr += phase_expr

            expr += ']'

        elif mode == 'latex':
            expr = 'e^{'

            if phase_expr != '1':
                expr += phase_expr

            if self.phase_norm is not None:
                if self.phase_norm[1] and self.phase_norm[1][0].isnumeric():
                    expr += r' \cdot '

                expr += self.phase_norm[1]

            expr += ' i}'
        
        return expr


class QPrintBraKet(QPrintBase):
    """Helper class to compose an expression of a given quantum object.

    Args:
        qobj: Input quantum object.
        amp_norm: Specification of the normalization of amplitudes by (numeric devisor, unit in LaTeX).
        phase_norm: Specification of the normalization of phases by (numeric devisor, unit in LaTeX).
        global_phase: Specification of the phase to factor out. Give a numeric offset or 'mean'.
        terms_per_row: Number of terms to show per row.
        amp_format: Format for the numerical value of the amplitude absolute values.
        phase_format: Format for the numerical value of the phases.
        epsilon: Numerical cutoff for ignoring amplitudes (relative to max) and phase (absolute).
        lhs_label: If not None, prepend 'label = ' to the printout.
        subsystem_dims: Specification of the dimensions of the subsystems.
        binary: Show bra and ket indices in binary.
    """
    @dataclass
    class Term(QPrintBase.Term):
        ket: Optional[str] = None
        bra: Optional[str] = None
        
    class QobjType(Enum):
        KET = 1
        BRA = 2
        OPER = 3
        
    @staticmethod
    def _type_and_dim(qobj):
        if len(qobj.shape) == 1 or qobj.shape[1] == 1:
            objtype = QPrintBraKet.QobjType.KET
            dim = qobj.shape[0]
        elif qobj.shape[0] == 1 and qobj.shape[1] != 1:
            objtype = QPrintBraKet.QobjType.BRA
            dim = qobj.shape[1]
        else:
            objtype = QPrintBraKet.QobjType.OPER
            dim = qobj.shape[0]
            
        return objtype, dim
    
    def __init__(
        self,
        qobj: Any,
        amp_norm: Optional[Union[Number, Tuple[Number, str]]] = None,
        phase_norm: Optional[Tuple[Number, str]] = (np.pi, 'π'),
        global_phase: Optional[Union[Number, str]] = None,
        terms_per_row: int = 0,
        amp_format: str = '.3f',
        phase_format: str = '.2f',
        epsilon: float = 1.e-6,
        lhs_label: Optional[str] = None,
        subsystem_dims: Optional[array_like] = None,
        binary: bool = False
    ):
        super().__init__(
            qobj=qobj,
            amp_norm=amp_norm,
            phase_norm=phase_norm,
            global_phase=global_phase,
            terms_per_row=terms_per_row,
            amp_format=amp_format,
            phase_format=phase_format,
            epsilon=epsilon,
            lhs_label=lhs_label)

        self.subsystem_dims = subsystem_dims
        self.binary = binary
        
    def _compose(self):
        global_sign, global_amp, global_phase, terms = self._process()
        
        qobj, _ = self._qobj_data()
        
        objtype, dim = self._type_and_dim(qobj)
        
        has_ket = objtype in (QPrintBraKet.QobjType.KET, QPrintBraKet.QobjType.OPER)
        has_bra = objtype in (QPrintBraKet.QobjType.BRA, QPrintBraKet.QobjType.OPER)
        
        subsystem_dims = self.subsystem_dims
        if subsystem_dims is None:
            subsystem_dims = np.array([dim])
            
        assert np.prod(subsystem_dims) == dim, (f'Product of subsystem dimensions {np.prod(subsystem_dims)}'
                                                f' and qobj dimension {dim} do not match')
        
        # State label format template
        if self.binary:
            log2_dims = np.log2(np.asarray(subsystem_dims))
            assert np.allclose(log2_dims, np.round(log2_dims))
            label_template = ','.join(f'{{:0{s}b}}' for s in log2_dims.astype(int))
        else:
            label_template = ','.join(['{}'] * len(subsystem_dims))

        # Make tuples of quantum state labels and format the term indices
        if has_scipy and isinstance(qobj, scipy.sparse.csr_matrix):
            # CSR matrix: diff if indptr = number of elements in each row
            repeats = np.diff(qobj.indptr)
            row_labels_flat = np.repeat(np.arange(qobj.shape[0]), repeats)
            # unravel into row indices accounting for the tensor product
            if has_ket:
                row_labels = np.unravel_index(row_labels_flat, subsystem_dims)
            if has_bra:
                col_labels = np.unravel_index(qobj.indices, subsystem_dims)
                
        elif isinstance(qobj, np.ndarray):
            if has_ket:
                row_labels = np.unravel_index(np.arange(dim), subsystem_dims)
            if has_bra:
                col_labels = np.unravel_index(np.arange(dim), subsystem_dims)

        # Update the term objects with the basis labels
        terms_with_labels = []
        
        for term in terms:
            term = QPrintBraKet.Term(index=term.index, sign=term.sign, amp=term.amp, phase=term.phase)
            
            if has_ket:
                term.ket = label_template.format(*(r[term.index[0]] for r in row_labels))
            if has_bra:
                # idx can be an 1- or 2-tuple depending on the type of self.qobj
                term.bra = label_template.format(*(c[term.index[-1]] for c in col_labels))
                
            terms_with_labels.append(term)

        return global_sign, global_amp, global_phase, terms_with_labels
    
    def _format_label(self, term, mode) -> str:
        expr = ''

        if mode == 'text':
            if term.ket is not None:
                expr += f'|{term.ket}>'
            if term.bra is not None:
                expr += f'<{term.bra}|'

        elif mode == 'latex':
            if term.ket is not None:
                expr += fr'| {term.ket} \rangle'
            if term.bra is not None:
                expr += fr'\langle {term.bra} |'            

        return expr
    
    def _format_lhs(self, mode) -> str:
        if not self.lhs_label:
            return ''
        
        objtype, _ = self._type_and_dim(self._qobj_data()[0])

        if mode == 'text':
            if objtype == QPrintBraKet.QobjType.KET:
                return f'|{self.lhs_label}>'
            elif objtype == QPrintBraKet.QobjType.BRA:
                return f'<{self.lhs_label}|'
        
        elif mode == 'latex':
            if objtype == QPrintBraKet.QobjType.KET:
                return fr'| {self.lhs_label} \rangle'
            elif objtype == QPrintBraKet.QobjType.BRA:
                return fr'\langle {self.lhs_label} |'

        return self.lhs_label


def qprint(
    qobj: Any,
    amp_norm: Optional[Union[Number, Tuple[Number, str]]] = None,
    phase_norm: Optional[Tuple[Number, str]] = (np.pi, 'π'),
    global_phase: Optional[Union[Number, str]] = None,
    terms_per_row: int = 0,
    amp_format: str = '.3f',
    phase_format: str = '.2f',
    epsilon: float = 1.e-6,
    lhs_label: Optional[str] = None,
    subsystem_dims: Optional[array_like] = None,
    binary: bool = False,
    fmt: Optional[str] = None
) -> PrintReturnType:
    """Pretty-print a quantum object.
    
    Three printing formats are supported:
    - `'text'`: Print text to stdout.
    - `'latex'`: Return an object processable by IPython into a typeset LaTeX expression.
    - `'mpl'`: Return a matplotlib figure.
    
    Args:
        qobj: Input quantum object.
        amp_norm: Specification of the normalization of amplitudes by (numeric devisor, unit in LaTeX).
        phase_norm: Specification of the normalization of phases by (numeric devisor, unit in LaTeX).
        global_phase: Specification of the phase to factor out. Give a numeric offset or 'mean'.
        terms_per_row: Number of terms to show per row.
        amp_format: Format for the numerical value of the amplitude absolute values.
        phase_format: Format for the numerical value of the phases.
        epsilon: Numerical cutoff for ignoring amplitudes (relative to max) and phase (absolute).
        lhs_label: If not None, prepend 'label = ' to the printout.
        subsystem_dims: Specification of the dimensions of the subsystems.
        binary: Show bra and ket indices in binary.
        fmt: Output format.
    """
    if fmt is None:
        if hasattr(builtins, '__IPYTHON__'):
            fmt = 'latex'
        else:
            fmt = 'text'
            
    pobj = QPrintBraKet(qobj=qobj,
                        amp_norm=amp_norm,
                        phase_norm=phase_norm,
                        global_phase=global_phase,
                        terms_per_row=terms_per_row,
                        amp_format=amp_format,
                        phase_format=phase_format,
                        epsilon=epsilon,
                        lhs_label=lhs_label,
                        subsystem_dims=subsystem_dims,
                        binary=binary)
            
    if fmt == 'text':
        return pobj
    elif fmt == 'latex':
        return LaTeXRepr(pobj)
    elif fmt == 'mpl':
        return pobj.mpl()
    else:
        raise NotImplementedError(f'qprint with format {fmt} not implemented')

        
class QPrintPauli(QPrintBase):
    """Helper class to compose an expression for a Pauli decomposition from a matrix or components.

    Args:
        qobj: A square matrix (shape `(d1*d2*..., d1*d2*...)`) or a components array
            (shape `(d1**2, d2**2, ...)`). Argument `dim` is required for the matrix interpretation.
        amp_norm: Specification of the normalization of amplitudes by (numeric devisor, unit in LaTeX).
        phase_norm: Specification of the normalization of phases by (numeric devisor, unit in LaTeX).
        global_phase: Specification of the phase to factor out. Give a numeric offset or 'mean'.
        terms_per_row: Number of terms to show per row.
        amp_format: Format for the numerical value of the amplitude absolute values.
        phase_format: Format for the numerical value of the phases.
        epsilon: Numerical cutoff for ignoring amplitudes (relative to max) and phase (absolute).
        lhs_label: If not None, prepend 'label = ' to the printout.
        dim: Specification of the dimensions of the subsystems. Used only when `qobj` is a square
            matrix or a 1D array.
        symbol: Pauli matrix symbols.
        delimiter: Pauli product delimiter.
    """
    @dataclass
    class Term(QPrintBase.Term):
        pauli: str
    
    def __init__(
        self,
        qobj: Any,
        amp_norm: Optional[Union[Number, Tuple[Number, str]]] = None,
        phase_norm: Optional[Tuple[Number, str]] = (np.pi, 'π'),
        global_phase: Optional[Union[Number, str]] = None,
        terms_per_row: int = 0,
        amp_format: str = '.3f',
        phase_format: str = '.2f',
        epsilon: float = 1.e-6,
        lhs_label: Optional[str] = None,
        dim: Optional[array_like] = None,
        symbol: Optional[Union[str, Sequence[str]]] = None,
        delimiter: str = ''
    ):
        super().__init__(
            qobj=qobj,
            amp_norm=amp_norm,
            phase_norm=phase_norm,
            global_phase=global_phase,
            terms_per_row=terms_per_row,
            amp_format=amp_format,
            phase_format=phase_format,
            epsilon=epsilon,
            lhs_label=lhs_label)

        self.dim = dim
        self.symbol = symbol
        self.delimiter = delimiter
        
    def _compose(self):
        qobj, _ = self._qobj_data(self.qobj)
        
        if self.dim is not None:
            if len(qobj.shape) == 2 and qobj.shape[0] == qobj.shape[1]:
                # This is a matrix -> extract the components
                try:
                    matrix = qobj.toarray()
                except AttributeError:
                    matrix = qobj

                qobj = paulis.components(matrix, dim=self.dim)

            elif len(qobj.shape) == 1:
                try:
                    components = qobj.toarray()
                except AttributeError:
                    components = qobj
                    
                qobj = components.reshape(np.square(self.dim))
        
        global_sign, global_amp, global_phase, terms = self._process(qobj)

        dim = np.around(np.sqrt(qobj.shape)).astype(int)
        
        labels = paulis.labels(dim, symbol=self.symbol, delimiter=self.delimiter)

        # Update the term objects with the basis labels
        terms_with_labels = []
        
        for term in terms:
            term = QPrintPauli.Term(index=term.index, sign=term.sign, amp=term.amp,
                                    phase=term.phase, pauli=labels[term.index])
            
            terms_with_labels.append(term)

        return global_sign, global_amp, global_phase, terms_with_labels
            
    def _format_label(self, term, mode):
        return term.pauli
    
    def _format_lhs(self, mode) -> str:
        return self.lhs_label


def pauliprint(
    qobj: Any,
    amp_norm: Optional[Union[Number, Tuple[Number, str]]] = None,
    phase_norm: Optional[Tuple[Number, str]] = (np.pi, 'π'),
    global_phase: Optional[Union[Number, str]] = None,
    terms_per_row: int = 0,
    amp_format: str = '.3f',
    phase_format: str = '.2f',
    epsilon: float = 1.e-6,
    lhs_label: Optional[str] = None,
    dim: Optional[array_like] = None,
    symbol: Optional[Union[str, Sequence[str]]] = None,
    delimiter: str = '',
    fmt: Optional[str] = None
) -> PrintReturnType:
    """Pretty-print a matrix or an array of Pauli components.

    Express the argument as a sum of Pauli strings. When a matrix is passed, it is Pauli-decomposed
    according to the `dim` parameter. 

    Three printing formats are supported:
    - `'text'`: Print text to stdout.
    - `'latex'`: Return an object processable by IPython into a typeset LaTeX expression.
    - `'mpl'`: Return a matplotlib figure.
    
    Args:
        qobj: Input quantum object.
        amp_norm: Specification of the normalization of amplitudes by (numeric devisor, unit in LaTeX).
        phase_norm: Specification of the normalization of phases by (numeric devisor, unit in LaTeX).
        global_phase: Specification of the phase to factor out. Give a numeric offset or 'mean'.
        terms_per_row: Number of terms to show per row.
        amp_format: Format for the numerical value of the amplitude absolute values.
        phase_format: Format for the numerical value of the phases.
        epsilon: Numerical cutoff for ignoring amplitudes (relative to max) and phase (absolute).
        lhs_label: If not None, prepend 'label = ' to the printout.
        dim: Specification of the dimensions of the subsystems. Used only when `qobj` is a square
            matrix or a 1D array.
        symbol: Pauli matrix symbols.
        delimiter: Pauli product delimiter.
    """
    if fmt is None:
        if hasattr(builtins, '__IPYTHON__'):
            fmt = 'latex'
        else:
            fmt = 'text'
            
    pobj = QPrintPauli(qobj=qobj,
                       amp_norm=amp_norm,
                       phase_norm=phase_norm,
                       global_phase=global_phase,
                       terms_per_row=terms_per_row,
                       amp_format=amp_format,
                       phase_format=phase_format,
                       epsilon=epsilon,
                       lhs_label=lhs_label,
                       dim=dim,
                       symbol=symbol,
                       delimiter=delimiter)
            
    if fmt == 'text':
        return pobj
    elif fmt == 'latex':
        return LaTeXRepr(pobj)
    elif fmt == 'mpl':
        return pobj.mpl()
    else:
        raise NotImplementedError(f'qprint with format {fmt} not implemented')



from typing import Union

import numpy as np
ndarray = np.ndarray
try:
    import jax.numpy as jnp
except ImportError:
    pass
else:
    ndarray = Union[ndarray, jnp.ndarray]

array_like = Union[list, tuple, ndarray]



"""
===================================================
Useful quantum math functions (:mod:`rqutils.math`)
===================================================

.. currentmodule:: rqutils.math

Math API
========

.. autosummary::
   :toctree: ../generated
   
   matrix_ufunc
   matrix_exp
   matrix_angle
"""

from typing import Callable, Tuple, Any, Union
import sys
import tempfile
import pickle
import numpy as np
try:
    import h5py
except ImportError:
    has_h5py = False
else:
    has_h5py = True

from ._types import ndarray, array_like

def matrix_ufunc(
    op: Callable,
    mat: array_like,
    hermitian: Union[int, bool] = 0,
    with_diagonals: bool = False,
    npmod=np,
    save_errors=False
) -> Union[ndarray, Tuple[ndarray, ndarray]]:
    """Apply a unitary-invariant unary matrix operator to an array of normal matrices.
    
    The argument `mat` must be an array of normal (i.e. square diagonalizable) matrices in the last
    two dimensions. This function unitary-diagonalizes the matrices, applies `op` to the diagonals,
    and inverts the diagonalization.
    
    Args:
        op: Unary operator to be applied to the diagonals of `mat`.
        mat: Array of normal matrices (shape (..., n, n)). No check on normality is performed.
        hermitian: 1 or True -> `mat` is Hermitian, -1 -> `mat` is anti-hermitian, 0 or False -> otherwise
        with_diagonals: If True, also return the array `op(eigenvalues)`.

    Returns:
        An array corresponding to `op(mat)`. If `diagonals==True`, another array corresponding to `op(eigvals)`.
    """
    try:
        if hermitian == 1 or hermitian == True:
            eigvals, eigcols = npmod.linalg.eigh(mat)
        elif hermitian == -1:
            eigvals, eigcols = npmod.linalg.eigh(1.j * mat)
            eigvals *= -1.j
        else:
            eigvals, eigcols = npmod.linalg.eig(mat)
    except:
        if save_errors:
            if has_h5py:
                with tempfile.NamedTemporaryFile(suffix='.h5', delete=False) as tmpf:
                    pass
    
                with h5py.File(tmpf.name, 'w') as out:
                    out.create_dataset('matrices', data=mat)
            else:
                with tempfile.NamedTemporaryFile(delete=False) as tmpf:
                    pickle.dump(mat, tmpf)
                
            sys.stderr.write(f'Error in eigendecomposition. Matrix saved at {tmpf.name}\n')
            
        raise
        
    eigrows = npmod.conjugate(npmod.moveaxis(eigcols, -2, -1))

    op_eigvals = op(eigvals)
    
    op_mat = npmod.matmul(eigcols * op_eigvals[..., None, :], eigrows)

    if with_diagonals:
        return op_mat, op_eigvals
    else:
        return op_mat

def matrix_exp(
    mat: array_like,
    hermitian: Union[int, bool] = 0,
    with_diagonals: bool = False,
    npmod=np,
    save_errors=False
) -> ndarray:
    """`matrix_ufunc(exp, ...)`"""
    return matrix_ufunc(npmod.exp, mat, hermitian=hermitian, with_diagonals=with_diagonals,
                        npmod=npmod, save_errors=save_errors)

def matrix_angle(
    mat: array_like,
    hermitian: Union[int, bool] = 0,
    with_diagonals: bool = False,
    npmod=np,
    save_errors=False
) -> ndarray:
    """`matrix_ufunc(angle, ...)`"""
    return matrix_ufunc(npmod.angle, mat, hermitian=hermitian, with_diagonals=with_diagonals,
                        npmod=npmod, save_errors=save_errors)




r"""
==================================================
Generalized Pauli matrices (:mod:`rqutils.paulis`)
==================================================

.. currentmodule:: rqutils.paulis

Fundamentals
============

Generalized :math:`n`-dimensional Pauli matrices :math:`\lambda^{(n)}_{k}` (:math:`0 \leq k \leq n^2 - 1`)
are defined recursively:

- :math:`\lambda^{(n)}_{0} = \sqrt{\frac{2}{n}} \mathrm{diag}(1, \dots, 1, 1)`
- :math:`\lambda^{(n)}_{k} = \mathrm{blkdiag}(\lambda^{(n-1)}_{k}, 0)` for :math:`1 \leq k < (n-1)^2`
- :math:`(\lambda^{(n)}_{(n-1)^2 + k})_{ab} = \xi_k \delta_{k//2, a}\delta_{n-1, b} + \eta_k \delta_{n-1, a}\delta_{k//2, b}`
  for :math:`0 \leq k < 2(n-1)`, with :math:`\xi_k = \eta_k = 1` (:math:`k` even) and :math:`-\xi_k = \eta_k = i` (:math:`k` odd)
- :math:`\lambda^{(n)}_{n^2-1} = \sqrt{\frac{2}{n(n-1)}} \mathrm{diag}(1, \dots, 1, -n+1)`

These matrices satisfy the normalization condition

.. math::

    \mathrm{tr}(\lambda^{(n)}_k \lambda^{(n)}_l) = 2 \delta_{k, l}

and thus form an orthonormal basis for the space of :math:`n`-dimensional Hermitian matrices. 

Implications of the normalization
---------------------------------

Any :math:`n`-dimensional Hermitian matrix :math:`H` can be decomposed into a form

.. math::

    H = \sum_{k=0}^{n^2-1} \nu_k \lambda^{(n)}_k.

To extract the coefficient :math:`\nu_k`, one needs to compute

.. math::

    \nu_k = \frac{1}{2} \mathrm{tr}(\lambda^{(n)}_k H),

i.e., divide the product trace by 2.

Also, note that :math:`\lambda^{(n)}_{0}` is *not* the :math:`n`-dimensional identity matrix but differ
from it by a factor :math:`\sqrt{\frac{2}{n}}`.


Pauli products
==============

A physical composite system of :math:`s` subsystems is usually better described in terms of a tensor
product of :math:`s` Hamiltonians each of dimension :math:`n_i (i=1, \dots, s)`, rather than a single
Hamiltonian of :math:`N := \prod_{i=1}^{s} n^i` dimensions. A natural decomposition of the former
would be in terms of tensor products of :math:`s` Pauli matrices

.. math::

    \Lambda^{(n_1 \dots n_s)}_{k_1 \dots k_s} = \frac{1}{2^{s-1}} \bigotimes_{i=1}^{s} \lambda^{(n_i)}_{k_i},

which constitute an orthonormal basis of the space of :math:`N`-dimensional Hermitian matrices with
a rather awkward normalization

.. math::

    \mathrm{tr}(\Lambda^{(n_1 \dots n_s)}_{k_1 \dots k_s} \Lambda^{(n_1 \dots n_s)}_{l_1 \dots l_s}) = 2 \frac{1}{2^{s-1}} \prod_i \delta_{k_i, l_i}.

The full `s`-body Hamiltonian :math:`H` is decomposed into

.. math::

    H = \sum_{k_1 \dots k_s} \nu_{k_1 \dots k_s} \Lambda^{(n_1 \dots n_s)}_{k_1 \dots k_s},

and the component :math:`\nu_{k_1 \dots k_s}` is extracted by

.. math::

    \nu_{k_1 \dots k_s} = 2^{s-2} \mathrm{tr}(\Lambda^{(n_1 \dots n_s)}_{k_1 \dots k_s} H).


Dimension truncation
====================

Thanks to the recursive definition of the Pauli matrices, the decomposition of an :math:`m`-dimensional
submatrix of an :math:`n`-dimensional Hermitian matrix is mostly trivially obtained from the components
of the latter:

If

.. math::

    H^{(n)} = \sum_{k=0}^{n^2-1} \nu_k \lambda^{(n)}_k

then the truncated matrix :math:`\bar{H}^{(m)}` is

.. math::

    \bar{H}^{(m)} = \sum_{k=0}^{m^2-1} \bar{\nu}_k \lambda^{m}_k,

with :math:`\bar{\nu}_k = \nu_k` for :math:`1 \leq k \leq m^2 - 1`. For :math:`k=0`, however, we need to
consider the projection of the diagonal matrices:

.. math::

    \mathrm{tr}_{m} (\lambda^{m}_0 \bar{\lambda}^{(n|m)}_0) & = 2 \sqrt{\frac{m}{n}}, \\
    \mathrm{tr}_{m} (\lambda^{m}_0 \bar{\lambda}^{(n|m)}_{d^2-1}) & = 2 \sqrt{\frac{m}{d(d-1)}} \quad \text{if} \; d > m

where :math:`\mathrm{tr}_{m}(\cdot)` represents the :math:`m`-dimensional trace, and :math:`\bar{\lambda}^{(n|m)}_k`
is the :math:`m`-dimensional submatrix of :math:`\lambda^{(n)}_k`. Thus we have

.. math::

    \bar{\nu}_0 = \frac{1}{2} \mathrm{tr}_m (\lambda^m_0 \bar{H}^{(m)}) = \sqrt{\frac{m}{n}} \nu_0 + \sum_{d > m} \sqrt{\frac{m}{d(d-1)}} \nu_{d^2-1}.
    
Pauli Matrices API
==================

.. autosummary::
   :toctree: ../generated
   
   paulis
   components
   compose
   l0_projector
   truncate
   symmetry
   labels
"""

from typing import Sequence, Optional, Union
from types import ModuleType
import string
import numpy as np
try:
    import jax
    import jax.numpy as jnp
except ImportError:
    has_jax = False
else:
    has_jax = True

from ._types import array_like, ndarray

def paulis(dim: Union[int, Sequence[int]]) -> np.ndarray:
    r"""Return a list of generalized Pauli matrices of given dimension(s) as an array.
 
    Args:
        dim: Dimension(s) of the Pauli matrices.
        
    Returns:
        An array of Pauli (product) matrices as an array. For `dim=(d1, d2, ...)`, the shape of
        the array is `(d1**2, d2**2, ..., d1*d2*..., d1*d2*...)`.
    """
    if isinstance(dim, int):
        dim = (dim,)
        
    subsystems = []
        
    for pauli_dim in dim:
        # Compose the unnormalized matrices
        matrices = np.zeros((pauli_dim ** 2, pauli_dim, pauli_dim), dtype=complex)
        matrices[0] = np.diag(np.ones(pauli_dim))
        ip = 1
        for isub in range(1, pauli_dim):
            for irow in range(isub):
                matrices[ip, irow, isub] = 1.
                matrices[ip, isub, irow] = 1.
                ip += 1
                matrices[ip, irow, isub] = -1.j
                matrices[ip, isub, irow] = 1.j
                ip += 1

            matrices[ip, :isub + 1, :isub + 1] = np.diag(np.array([1.] * isub + [-isub]))
            ip += 1

        # Normalization
        norm = np.trace(np.matmul(matrices, matrices), axis1=1, axis2=2)
        matrices *= np.sqrt(2. / norm)[:, None, None]

        subsystems.append(matrices)
        
    num_sub = len(dim)

    if num_sub == 1:
        return subsystems[0]
    
    else:
        # Compose Pauli products
        # (d1**2, d1, d1) x (d2**2, d2, d2) -> (d1**2, d2**2, d1*d2, d1*d2)
        #      a   b   c         d   e   f          a      d     be     cf
        # be and cf are reshaped into 1 dimension each
        chars = string.ascii_letters
        if num_sub * 3 > len(chars):
            raise NotImplemented('Too many subsystems - need an implementation using recursive np.kron')

        indices_in = []
        indices_out = [''] * 3
        for il in range(0, num_sub * 3, 3):
            indices_in.append(chars[il:il + 3])
            indices_out[0] += chars[il]
            indices_out[1] += chars[il + 1]
            indices_out[2] += chars[il + 2]

        indices = f'{",".join(indices_in)}->{"".join(indices_out)}'
        dim_array = np.asarray(dim)
        shape = np.concatenate((np.square(dim_array), np.prod(np.repeat(dim_array[None, :], 2, axis=0), axis=1)))
        
        return np.einsum(indices, *subsystems).reshape(*shape) / (2 ** (num_sub - 1))


def components(
    matrix: array_like,
    dim: Optional[Sequence[int]] = None,
    npmod: ModuleType = np
) -> ndarray:
    r"""Return the Pauli decomposition coefficients :math:`\nu_{k_1 \dots k_n}` of the matrix.
    
    Args:
        matrix: Matrix to decompose. The last two dimensions of the array are dotted with the Pauli
            matrices.
        dim: Subsystem dimensions. The product of subsystem dimensions must match the matrix dimension.
            If None, the matrix is assumed to represent a single system.
        
    Returns:
        A complex array of shape `(..., d1**2, d2**2, ...)` where `d1`, `d2`, ... are the subsystem dimensions.
        
    Raises:
        ValueError: If `prod(dim)` does not match the matrix dimension.
    """
    if npmod is np:
        if dim is None:
            dim = (matrix.shape[-1],)
        elif np.prod(dim) != matrix.shape[-1]:
            raise ValueError(f'Invalid subsystem dimensions {dim}')
        
    basis = paulis(dim)

    return npmod.tensordot(matrix, basis, ((-2, -1), (-1, -2))) * (2 ** (len(dim) - 2))


def compose(
    components: array_like,
    dim: Optional[Sequence[int]] = None,
    npmod: ModuleType = np
) -> ndarray:
    r"""Compose a matrix from the Pauli components.
    
    Args:
        components: Pauli components of the desired matrix, shape (..., d1**2, d2**2, ...)
        dim: Subsystem dimensions. If present, last `len(dim)` dimensions of `components`
            are dotted with the corresponding Pauli matrices.
        
    Returns:
        A complex array of shape `(..., d1*d2*..., d1*d2*...)`.
    """
    if dim is None:
        dim = np.around(np.sqrt(components.shape)).astype(int)

        if npmod is np and not np.allclose(np.square(dim), components.shape):
            raise ValueError('Components array shape invalid')

    basis = paulis(dim)

    comp_axes = list(range(-len(dim), 0))
    pauli_axes = list(range(len(dim)))
    return npmod.tensordot(components, basis, (comp_axes, pauli_axes))
    
    
def l0_projector(reduced_dim: int, original_dim: int) -> np.ndarray:
    r"""Return the vector that projects the components of original_dim decomposition onto lambda_0 of reduced_dim.
    
    Args:
        reduced_dim: Matrix dimension of the target subspace.
        original_dim: Matrix dimension of the full space.
        
    Returns:
        Projection vector :math:`\vec{v}` that gives :math:`\bar{\nu}_0 = \vec{v} \cdot \vec{\nu}`.
    """
    if reduced_dim > original_dim:
        raise ValueError('Reduced dim greater than original dim')
    
    projector = np.zeros(original_dim ** 2)
    projector[0] = np.sqrt(reduced_dim / original_dim)
    
    for d in range(reduced_dim + 1, original_dim + 1):
        projector[d ** 2 - 1] = np.sqrt(reduced_dim / d / (d - 1))

    return projector


def truncate(
    components: array_like,
    reduced_dim: Union[int, Sequence[int]],
    npmod: ModuleType = np
) -> ndarray:
    r"""Truncate a component array of a matrix into the components for a submatrix.
    
    The component array can have extra dimensions in front (e.g. time axis if this is a time series of components).
    In such a case, reduced_dim must be a sequence of integers with the length correpsonding to the number of
    subsystems.
    
    Args:
        components: Pauli components of the original matrix, shape (..., d1**2, d2**2, ...)
        reduced_dim: Dimension(s) of the submatrix(es).
        
    Returns:
        Components of the submatrix, shape (..., r1**2, r2**2, ...)
    """
    if npmod is np and isinstance(reduced_dim, int):
        reduced_dim = (reduced_dim,) * len(components.shape)
        
    num_subsystems = len(reduced_dim)

    original_shape = components.shape[-num_subsystems:]
    reduced_shape = np.square(reduced_dim)

    if npmod is np:
        if np.any(reduced_shape > np.asarray(original_shape)):
            raise ValueError(f'Reduced dimensions greater than original dimensions')
            
        if np.allclose(reduced_shape, original_shape):
            return components.copy()
        
    original_dim = npmod.around(npmod.sqrt(original_shape)).astype(int)

    def project_dim(idim, components):
        # Construct a matrix (v, diag([1] * reduced)[1:], diag([0] * truncated))^T 
        projector = l0_projector(reduced_dim[idim], original_dim[idim])
        diag = npmod.concatenate((npmod.ones(reduced_dim[idim] ** 2),
                                  npmod.zeros(original_dim[idim] ** 2 - reduced_dim[idim] ** 2)))
        projector = npmod.concatenate((projector[None, :], npmod.diag(diag)[1:]), axis=0)

        return npmod.tensordot(projector, components, (1, -num_subsystems + idim))

    if has_jax and npmod is jnp:
        def loop_body(idim, components):
            return jax.lax.cond(reduced_dim[idim] == original_dim[idim],
                                lambda c: c,
                                lambda c: project_dim(idim, c),
                                components)
        
        components = jax.lax.fori_loop(0, num_subsystems, loop_body, components)
        
    else:
        for idim in range(num_subsystems):
            if reduced_dim[idim] != original_dim[idim]:
                components = project_dim(idim, components)

    slices = tuple(slice(shape) for shape in reduced_shape)

    return components[(...,) + slices]


def symmetry(dim: int):
    r"""Return the symmetry (-1, 0, 1) of the Pauli matrices.
    
    Args:
        dim: Dimension of the Pauli matrices.
        
    Returns:
        An 1D integer array with entries -1, 0, 1 depending on whether the corresponding Pauli matrix
        is antisymmetric, diagonal, or symmetric.
    """
    
    symmetry = np.zeros(dim ** 2, dtype=int)

    ip = 1
    for isub in range(1, dim):
        for irow in range(isub):
            symmetry[ip] = 1
            ip += 1
            symmetry[ip] = -1
            ip += 1

        ip += 1
        
    return symmetry


def labels(
    dim: Union[int, Sequence[int]],
    symbol: Optional[Union[str, Sequence[str]]] = None,
    delimiter: str = ''
) -> np.ndarray:
    r"""Generate the labels for the Pauli matrices of a given dimension.
    
    Args:
        dim: Dimension(s) of the Pauli matrices.
        symbol: Base symbol.
        delimiter: Delimiter between the symbols for multibody labels.
        
    Returns:
        An ndarray of type string and shape `(d1**2, d2**2, ...)`.
    """
    if isinstance(dim, int):
        dim = (dim,)
        
    if symbol is None or isinstance(symbol, str):
        symbol = (symbol,) * len(dim)
        
    out = np.array('', dtype=str)

    for pauli_dim, sym in zip(dim, symbol):
        if delimiter and len(out.shape) > 0:
            out = np.char.add(out, np.full_like(out, delimiter))
            
        if not sym:
            if pauli_dim == 2:
                labels = ['I', 'X', 'Y', 'Z']
            elif sym is None:
                labels = list(fr'\lambda_{i}' for i in range(pauli_dim ** 2))
            else:
                labels = list(str(i) for i in range(pauli_dim ** 2))
        else:
            labels = list(f'{sym}_{i}' for i in range(pauli_dim ** 2))
            
        out = np.char.add(np.repeat(out[..., None], pauli_dim ** 2, axis=-1), labels)
        
    return out
